#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include "pref.h"
#include "neal_matrix.h"

// matrix product with matrix, Cmk = Amn * Bnk 
void brmul(double a[],double b[],int m,int n,int k,double c[]) 
{
	int i,j,l,u;
    for (i=0; i<=m-1; i++)
    for (j=0; j<=k-1; j++)
      { u=i*k+j; c[u]=0.0;
        for (l=0; l<=n-1; l++)
          c[u]=c[u]+a[i*n+l]*b[l*k+j];
      }
    return;
}

/*
void sbrmul(double a[],double b[],int n,double c[]) 
{
	int i,j,l,u;
    for (i=0; i<=n-1; i++)
	{	
		for (j=0; j<=n-1; j++)
			c[u]=0.0 ;
	}
    for (i=0; i<=n-1; i++)
	{
		for (j=i+1; j<=n-1; j++)
		{ 
			c[i*n+j] += a[i*n+j]*b[j*n+i] ; 
			c[j*n+i] += a[i*n+j]*b[j*n+i] ;
		}
	}
	return;
}
*/

int Labels_LAPLACE_Evaluate_FuncGrad ( void * pointer, bfgs_FuncGrad * bfgs ) 
{
	Pref_Settings * settings ;
	Pref_Node * pref ;
	Data_Node * node ;
	Alphas * alpha ;
	Alphas * alphab ;
	double * w ;
	double * ph ;

	double * lambda ;
	double * dsigma ;
	double * dmat ;
	double * df ;
	double * dfk ;
	double * dls ;
	double * ssmat ;
	double * sssmat ;
	double * sdmat ;

	unsigned int i, j, k ;
	double sk ;

	double * t1 ;
	double * t2 ;
	double temp ;
	double cbe ;
	unsigned int ii=0 ; // settings->kappa	

	double sigma, z, n, phi ;

	if (NULL == pointer || NULL == bfgs)
		return 1 ;
	settings = (Pref_Settings *)pointer ;

	Bfgs_Pref_Settings( settings,settings->bfgs ) ;

	// call EP routine to find posterior distribution 
	if (1 == Labels_MAP_Training (settings) )
	{
		printf("MAP failed to get equilibrium.\n") ;
		return 1 ;
	}

	assert(NULL != settings->invcov) ;

	// evaluate functional and gradient
	
	lambda = (double*)calloc(settings->pairs->count*settings->pairs->count*
		settings->pairs->classes*settings->pairs->classes,sizeof(double)) ;	
	dmat = (double*)calloc(settings->pairs->count*settings->pairs->count*
		settings->pairs->classes*settings->pairs->classes,sizeof(double)) ;
	t1 = (double*)malloc(settings->pairs->count*settings->pairs->classes*sizeof(double)) ;
	t2 = (double*)malloc(settings->pairs->count*settings->pairs->classes*sizeof(double)) ;	
	ph = (double*)calloc(settings->pairs->count*settings->pairs->classes,sizeof(double)) ;		
	dls = (double*)calloc(settings->pairs->count*settings->pairs->count*
		settings->pairs->classes*settings->pairs->classes,sizeof(double)) ;
	df = (double*)malloc(settings->pairs->count*settings->pairs->classes*sizeof(double)) ;	
	dsigma = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	ssmat = (double*)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ;
	sssmat = (double*)calloc(settings->pairs->count*settings->pairs->count*
		settings->pairs->classes*settings->pairs->classes,sizeof(double)) ;
	sdmat = (double*)malloc(settings->pairs->count*settings->pairs->count*
		settings->pairs->classes*settings->pairs->classes*sizeof(double)) ;
	w = (double*)calloc(settings->pairs->count*settings->pairs->classes,sizeof(double)) ;
	dfk = (double*)malloc(settings->pairs->count*settings->pairs->classes*sizeof(double)) ;	


	sigma = sqrt(2.0*settings->noisevar) ;

	bfgs->fx = settings->trace*settings->pairs->classes ;
	for (i=0;i<settings->number;i++)
		bfgs->gx[i] = 0 ;

	// fx = 0.5*w*Sigma*w
	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<i;j++)
		{
			temp = 0 ;
			for (k=0;k<settings->pairs->classes;k++)
				temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+j)->pair->weight[k] ;
			bfgs->fx += Calculate_Covfun(settings->alpha+i,settings->alpha+j,settings)*temp ;
		}
		temp = 0 ;
		for (k=0;k<settings->pairs->classes;k++)
			temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+i)->pair->weight[k] ;
		bfgs->fx += 0.5 * Calculate_Covfun(settings->alpha+i,settings->alpha+i,settings)*temp ;
	}
	/*for (k=0;k<settings->pairs->classes;k++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<settings->pairs->count;j++)
				bfgs->fx += 0.5* settings->invcov[i*settings->pairs->count+j]*(settings->alpha+i)->p_cache[k]*(settings->alpha+j)->p_cache[k] ;
		}
	}*/
	// initialize Lambda
	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		pref = node->pref.front ;
		while (NULL != pref)
		{
			z = pref->z  ;
			n = pref->n ;
			phi = pref->phi ;

			w[(pref->u-1)*settings->pairs->count+i]+=n/phi/sigma ;		
			w[(pref->v-1)*settings->pairs->count+i]-=n/phi/sigma ;

			sk = (z*z*n/phi+z*n*n/phi*phi-n/phi)/settings->noisevar/sqrt(2.0) ;
			ph[(pref->u-1)*settings->pairs->count+i]+=sk ;
			ph[(pref->v-1)*settings->pairs->count+i]-=sk ;
			sk = (n/phi-2.0*(n/phi)*(n/phi)*(n/phi)-3.0*z*(n/phi)*(n/phi)-z*z*n/phi)/sigma/sigma/sigma ;
			
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] -= sk*sqrt(2.0)*z ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] -= sk*sqrt(2.0)*z ;				
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += sk*sqrt(2.0)*z ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += sk*sqrt(2.0)*z ; 
			
			lambda[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				( (n*n)/(phi*phi)+z*n/phi)/sigma/sigma ;
			lambda[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				( (n*n)/(phi*phi)+z*n/phi)/sigma/sigma ;
			lambda[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] -= 
				( (n*n)/(phi*phi)+z*n/phi)/sigma/sigma ;
			lambda[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] -= 
				( (n*n)/(phi*phi)+z*n/phi)/sigma/sigma ;

			bfgs->fx -= log(phi) ;		
			bfgs->gx[0] += z*n/phi/sigma ;			
			pref = pref -> next ;
		}
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;

	// Sigma^{-1}+Lambda
			
	for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
	{
		for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
		{
			//assert(lambda[i*settings->pairs->count*settings->pairs->classes+j]==lambda[j*settings->pairs->count*settings->pairs->classes+i]) ;
			dmat[i*settings->pairs->count*settings->pairs->classes+j] =
				lambda[i*settings->pairs->count*settings->pairs->classes+j] ;
		}
	}

	for (k=0;k<settings->pairs->classes;k++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<settings->pairs->count;j++)
			{
				assert(settings->invcov[i*settings->pairs->count+j]==settings->invcov[j*settings->pairs->count+i]) ;
				dmat[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
					+i*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+j] 
					+= settings->invcov[i*settings->pairs->count+j] ; 
			}
		}
	}
//		for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
//			for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
//				assert(dmat[i*settings->pairs->count*settings->pairs->classes+j]==dmat[j*settings->pairs->count*settings->pairs->classes+i]) ;

	if (0 == cholesky(dmat,settings->pairs->count*settings->pairs->classes,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	
	for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
		bfgs->fx += log(dmat[i*settings->pairs->count*settings->pairs->classes+i]) ;
	// end of fx

	if (0 == inverse_from_cholesky(dmat, t1, t2, settings->pairs->count*settings->pairs->classes) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	// d lambda / d sigma
	for (i=0;i<settings->pairs->classes*settings->pairs->count;i++)
	{
		for (j=0;j<i;j++)
		{
			assert(lambda[i*settings->pairs->count*settings->pairs->classes+j]==lambda[j*settings->pairs->count*settings->pairs->classes+i]) ;
			dls[i*settings->pairs->classes*settings->pairs->count+j] -= 
				2.0*lambda[i*settings->pairs->classes*settings->pairs->count+j]/sigma ;
			dls[j*settings->pairs->classes*settings->pairs->count+i] -= 
				2.0*lambda[j*settings->pairs->classes*settings->pairs->count+i]/sigma ;
		}
		dls[i*settings->pairs->classes*settings->pairs->count+i] -= 
			2.0*lambda[i*settings->pairs->classes*settings->pairs->count+i]/sigma ;
	}

	// for gx[0] --- ln\sigma
	// df/ds
	for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)	
	{
		df[i] = 0 ;
		for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
		{
			assert(dls[i*settings->pairs->count*settings->pairs->classes+j]==dls[j*settings->pairs->count*settings->pairs->classes+i]) ;
			df[i]+=dmat[i*settings->pairs->count*settings->pairs->classes+j]*ph[j] ;
		}
	}

	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		pref = node->pref.front ;
		while (NULL != pref)
		{
			n=pref->n ;
			phi=pref->phi ;
			z=pref->z ;
			sk = (n/phi-2.0*(n/phi)*(n/phi)*(n/phi)-3.0*z*(n/phi)*(n/phi)-z*z*n/phi)/sigma/sigma/sigma ;

			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(df[(pref->u-1)*settings->pairs->count+i]-df[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(df[(pref->u-1)*settings->pairs->count+i]-df[(pref->v-1)*settings->pairs->count+i]) ;				
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(-df[(pref->u-1)*settings->pairs->count+i]+df[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(-df[(pref->u-1)*settings->pairs->count+i]+df[(pref->v-1)*settings->pairs->count+i]) ;
			pref = pref -> next ;
		}
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;

	for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
	{
		for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
		{
			//assert(dls[i*settings->pairs->count*settings->pairs->classes+j]==dls[j*settings->pairs->count*settings->pairs->classes+i]) ;
			bfgs->gx[0] += 0.5*dmat[i*settings->pairs->count*settings->pairs->classes+j]
				*dls[j*settings->pairs->count*settings->pairs->classes+i] ;
		}
	}
	bfgs->gx[0] = bfgs->gx[0]*(-0.5)*sigma*sigma*sigma ; // d /d 1/sigma^2
	if (settings->kernel==LINEAR&&settings->ardon==1)
		bfgs->gx[0] = 0 ;

	// end of gx[0] ;

	
	// d ln kappa
	// for kernel parameters
	// calculate for each settings->kappa Gaussian parameter
	if (1 == settings->ardon)
	{
		for (ii=0;ii<settings->pairs->dimen;ii++)
		{			
			// calculate dsigmam for each settings->kappa parameter
			for (i=0;i<settings->pairs->count;i++)
			{
				alpha = settings->alpha + i ;
					for (j=0;j<=i;j++)
					{
						alphab = settings->alpha + j ;
							if ( GAUSSIAN == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->kernels[j]-settings->kappa_m)*(-(alpha->pair->
									point[ii]-alphab->pair->point[ii])*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0/settings->pairs->dimen) ;
							}
							else if ( LINEAR == settings->kernel )
							{
								dsigma[i*settings->pairs->count+j]=(alpha->pair->point[ii]*alphab->pair->point[ii])*settings->kappa_o ;
							}
							else if ( USERDEFINED == settings->kernel )
							{
								cbe = 0 ; // temp storage
								if (0==settings->pairs->featuretype[ii])
									cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii]) 
									*(alpha->pair->point[ii]-alphab->pair->point[ii])/2.0 ;
								else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
									cbe-=1.0/2.0 ;
								dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe ;
							}
							else
								printf ("Error in kernel type.\n") ;
							dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
					}											
			}
			brmul(settings->invcov, dsigma,  settings->pairs->count, settings->pairs->count, settings->pairs->count, ssmat) ;// VC*VC*a'*dsigma*a/2
			// create sssmat
			for (k=0;k<settings->pairs->classes;k++)
			{
				for (i=0;i<settings->pairs->count;i++)
					for (j=0;j<settings->pairs->count;j++)
						sssmat[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
						+i*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+j] 
						= ssmat[i*settings->pairs->count+j] ;
			}
			brmul(dmat, sssmat, settings->pairs->count*settings->pairs->classes, 
				settings->pairs->count*settings->pairs->classes, 
				settings->pairs->count*settings->pairs->classes, sdmat) ;

			// VC*VC*a'*dsigma*a/2
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<i;j++)
				{
					temp = 0 ;
					for (k=0;k<settings->pairs->classes;k++)
						temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+j)->pair->weight[k] ;
					bfgs->gx[ii+1] -= dsigma[i*settings->pairs->count+j]*temp ;
				}
				temp = 0 ;
				for (k=0;k<settings->pairs->classes;k++)
					temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+i)->pair->weight[k] ;
				bfgs->gx[ii+1] -= 0.5 * dsigma[i*settings->pairs->count+i]*temp ;
			}

			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
					bfgs->gx[ii+1] += 0.5*sdmat[i*settings->pairs->count*settings->pairs->classes+j]
						*lambda[j*settings->pairs->count*settings->pairs->classes+i] ;
			//dfk
			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
			{
				dfk[i] = 0 ;
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
				{
					dfk[i] += sdmat[i*settings->pairs->count*settings->pairs->classes+j]*w[j] ;
					dls[i*settings->pairs->count*settings->pairs->classes+j] = 0 ;
				} 
			}
//
	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		pref = node->pref.front ;
		while (NULL != pref)
		{
			n=pref->n ;
			phi=pref->phi ;
			z=pref->z ;
			sk = (n/phi-2.0*(n/phi)*(n/phi)*(n/phi)-3.0*z*(n/phi)*(n/phi)-z*z*n/phi)/sigma/sigma/sigma ;
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(dfk[(pref->u-1)*settings->pairs->count+i]-dfk[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(dfk[(pref->u-1)*settings->pairs->count+i]-dfk[(pref->v-1)*settings->pairs->count+i]) ;				
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(-dfk[(pref->u-1)*settings->pairs->count+i]+dfk[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(-dfk[(pref->u-1)*settings->pairs->count+i]+dfk[(pref->v-1)*settings->pairs->count+i]) ;
			pref = pref -> next ;
		}
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;

			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
					bfgs->gx[ii+1] += 0.5*dmat[i*settings->pairs->count*settings->pairs->classes+j]*dls[j*settings->pairs->count*settings->pairs->classes+i] ;

bfgs->gx[ii+1] += settings->regular*settings->kappa[ii] ;
bfgs->fx += 0.5*settings->regular*settings->kappa[ii]*settings->kappa[ii] ;

//bfgs->gx[ii+settings->pairs->classes] += settings->regular ;
//bfgs->fx += settings->regular*settings->kappa[ii] ;

			bfgs->gx[ii+1] = bfgs->gx[ii+1] * settings->kappa[ii] ;
		}
	}
        else if (settings->kernel!=LINEAR)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			alpha = settings->alpha + i ;
			for (j=0;j<=i;j++)
			{
				alphab = settings->alpha + j ;
				if ( GAUSSIAN == settings->kernel )
				{
					cbe = 0 ; // temp storage
					for (ii=0;ii<settings->pairs->dimen;ii++)
						cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
						*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
					dsigma[i*settings->pairs->count+j] = 
						(alpha->kernels[j]-settings->kappa_m)*cbe/2.0/settings->pairs->dimen ;
				} 
				else if (LINEAR == settings->kernel )
				{
					cbe = 0 ; // temp storage
					for (ii=0;ii<settings->pairs->dimen;ii++)
						cbe += alpha->pair->point[ii]*alphab->pair->point[ii] ;
					dsigma[i*settings->pairs->count+j] = cbe*settings->kappa_o ;
				}
				else if ( USERDEFINED == settings->kernel )
				{
					cbe = 0 ; // temp storage
					for (ii=0;ii<settings->pairs->dimen;ii++)
					{
						if (0==settings->pairs->featuretype[ii])
							cbe-=(alpha->pair->point[ii]-alphab->pair->point[ii])
							*(alpha->pair->point[ii]-alphab->pair->point[ii]) ;
						else if (alpha->pair->point[ii]!=alphab->pair->point[ii])
							cbe-=1.0 ;
					}
					dsigma[i*settings->pairs->count+j] = (alpha->kernels[j]-settings->kappa_m) * cbe / 2.0 ;
				}
				else
					printf ("Error in kernel type.\n") ;
				dsigma[j*settings->pairs->count+i] = dsigma[i*settings->pairs->count+j] ;
			}
		}

		brmul(settings->invcov, dsigma,  settings->pairs->count, settings->pairs->count, settings->pairs->count, ssmat) ;// VC*VC*a'*dsigma*a/2
		// create sssmat
			for (k=0;k<settings->pairs->classes;k++)
			{
				for (i=0;i<settings->pairs->count;i++)
				{
					for (j=0;j<settings->pairs->count;j++)
					{
						assert(dsigma[j*settings->pairs->count+i]==dsigma[i*settings->pairs->count+j]) ;
						assert(settings->invcov[j*settings->pairs->count+i]==settings->invcov[i*settings->pairs->count+j]) ;
						//assert(ssmat[j*settings->pairs->count+i]==ssmat[i*settings->pairs->count+j]) ;
						sssmat[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
						+i*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+j] 
						= ssmat[i*settings->pairs->count+j] ;
					}
				}
			}
		brmul(dmat, sssmat, settings->pairs->count*settings->pairs->classes, 
			settings->pairs->count*settings->pairs->classes, 
			settings->pairs->count*settings->pairs->classes, sdmat) ;

			// VC*VC*a'*dsigma*a/2
			for (i=0;i<settings->pairs->count;i++)
			{
				for (j=0;j<i;j++)
				{
					temp = 0 ;
					for (k=0;k<settings->pairs->classes;k++)
						temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+j)->pair->weight[k] ;
					bfgs->gx[1] -= dsigma[i*settings->pairs->count+j]*temp ;
				}
				temp = 0 ;
				for (k=0;k<settings->pairs->classes;k++)
					temp += (settings->alpha+i)->pair->weight[k]*(settings->alpha+i)->pair->weight[k] ;
				bfgs->gx[1] -= 0.5 * dsigma[i*settings->pairs->count+i]*temp ;
			}

			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
					bfgs->gx[1] += 0.5*sdmat[i*settings->pairs->count*settings->pairs->classes+j]
					*lambda[j*settings->pairs->count*settings->pairs->classes+i] ;
			//dfk
			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
			{
				dfk[i] = 0 ;
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
				{
					//assert(sdmat[j*settings->pairs->count+i]==sdmat[i*settings->pairs->count+j]) ;
					dfk[i] += sdmat[i*settings->pairs->count*settings->pairs->classes+j]*w[j] ;
					dls[i*settings->pairs->count*settings->pairs->classes+j] = 0 ; 
				}
			}

	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		pref = node->pref.front ;
		while (NULL != pref)
		{
			n=pref->n ;
			phi=pref->phi ;
			z=pref->z ;
			sk = (n/phi-2.0*(n/phi)*(n/phi)*(n/phi)-3.0*z*(n/phi)*(n/phi)-z*z*n/phi)/sigma/sigma/sigma ;
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(dfk[(pref->u-1)*settings->pairs->count+i]-dfk[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(dfk[(pref->u-1)*settings->pairs->count+i]-dfk[(pref->v-1)*settings->pairs->count+i]) ;				
			dls[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				sk*(-dfk[(pref->u-1)*settings->pairs->count+i]+dfk[(pref->v-1)*settings->pairs->count+i]) ;			
			dls[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				sk*(-dfk[(pref->u-1)*settings->pairs->count+i]+dfk[(pref->v-1)*settings->pairs->count+i]) ;
			pref = pref -> next ;
		}
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;

			for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
				for (j=0;j<settings->pairs->count*settings->pairs->classes;j++)
					bfgs->gx[1] += 0.5*dmat[i*settings->pairs->count*settings->pairs->classes+j]*dls[j*settings->pairs->count*settings->pairs->classes+i] ;

			bfgs->gx[1] = bfgs->gx[1] * settings->kappa_a ;

	}

#ifdef _ORDINAL_DEBUG
	//if (bfgs->x[0]>=100)
	//bfgs->gx[settings->number-1] = 0 ;
	printf("functional %f \n", bfgs->fx) ;
	for (i=0;i<settings->number;i++)
	{
		printf("gardient %f --- x %f\n", bfgs->gx[i], bfgs->x[i]) ;
	}
#endif 
	free(dls) ;
	free(dfk) ;
	free(df) ;
	free(ph) ;
	free(w) ;
	free(sssmat) ;
	free(ssmat) ;
	free(sdmat) ;	
	free(dmat) ;
	free(t2) ;
	free(t1) ;
	free(lambda) ;
	free(dsigma) ;
	return 0 ;
}

int	Labels_LAPLACE_Training (Pref_Settings * settings) 
{
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	printf("Laplacian Approximation around MAP estimate ...\n") ;

	if ( 1 == bfgs_minfunc ( Labels_LAPLACE_Evaluate_FuncGrad, settings, settings->bfgs ) )
	{
		printf("fail to find local minimum.\n") ;
		return 1;
	}
	// save best funcgrad to settings 
	Bfgs_Pref_Settings( settings, settings->bfgs ) ;
	printf("Optimal Settings: \n") ;
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	if (0==settings->ardon)		
		printf("Kernel: %f\n",settings->kappa_a) ;
	else if (settings->pairs->dimen<10) 
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %f:\n", i, settings->kappa[i]) ;
	}
	tstart();
	Labels_MAP_Training (settings) ;
	tend();
	settings->time = tval() ;
	return 0 ;
}

