#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <assert.h>
#include "pref.h"
#include "neal_matrix.h"

extern double lerrf(double) ;
extern double fromb(double mean, double sig, double eps) ;

int Pref_Prediction (Pref_Settings * settings) 
{
	Pref_Node * pref ;
	Data_List testdatalist ;
	Data_Node * node ;
	double * kcvx ;
	unsigned int i, k;
	char buf[2048] ;
	FILE * fid ;
	unsigned int error = 0, abserr = 0, count = 0, flag ;

	assert(NULL != settings) ;

	// compute the invcov
	assert(NULL != settings->invcov) ;

	// loading test data
	Create_Data_List (&testdatalist) ;
	if ( 1 == Labels_Loadfile (&testdatalist, settings->testfile) )
	{
		printf("fail to load data from %s.\n\n", settings->testfile) ;
		exit(1) ;
	}

	//sigma =  sqrt(2.0*settings->noisevar) ;
	sprintf(buf,"%s.guess",settings->testfile) ;
	fid = fopen(buf,"w+t") ;
	
	// compute predictions
	node = testdatalist.front ;
	while (NULL != node)
	{
		// for each point
		// cache kcx 
		node->weight = (double*)calloc(settings->pairs->classes,sizeof(double)) ;
		kcvx = (double*)malloc(settings->pairs->count*sizeof(double)) ;
		for (i=0;i<settings->pairs->count;i++)
		{
			kcvx[i] = Calc_Covfun((settings->alpha+i)->pair->point,node->point,settings) ;
		}
		for (k=0;k<settings->pairs->classes;k++)
		{
			for (i=0;i<settings->pairs->count;i++)
				node->weight[k] += kcvx[i]*(settings->alpha+i)->pair->weight[k] ;
		}
		free(kcvx) ;
		flag = 0 ;
		pref = node->pref.front ;
		while(NULL != pref)
		{
			count += 1 ;
			if (node->weight[pref->u-1]<=node->weight[pref->v-1])
			{
				flag += 1 ;
				error += 1 ;
			}
			if (NULL != fid)
			{
				fprintf(fid,"%d ",node->weight[pref->u-1]>node->weight[pref->v-1]?0:1) ;
			}
			pref = pref->next ;
		}
		if (NULL != fid) 
			fprintf(fid,"\n") ;
		if (flag>0)
			abserr += 1 ;
		node=node->next ;
	}
	if (NULL != fid)
		fclose(fid) ;

	// save the function values
	sprintf(buf,"%s.func",settings->testfile) ;
	fid = fopen(buf,"w+t") ;
	if (NULL != fid)
	{
		node = testdatalist.front ;
		while (NULL != node)
		{
			for (k=0;k<settings->pairs->classes;k++)
				fprintf(fid, "%.4f ", node->weight[k]) ;
			fprintf(fid, "\n") ;
			node=node->next ;
		}
		fclose(fid) ;
	}
	
	printf("Testing preference error number %u/%u with sample error %u/%lu.\n", error, count, abserr, testdatalist.count) ;

	fid = fopen ("gplabels_lap_batch.log","a+t") ;
	if (NULL != fid)
	{
		fprintf(fid,"%u %f %u %f %f\n", error, (double)error/(double)count, abserr, (double)abserr/(double)(testdatalist.count), settings->time) ;
		fclose(fid) ;
	}
	// save log files
	Clear_Data_List (&testdatalist) ;
	return 0 ;
}

int Dumping_Pref_Settings (Pref_Settings * settings)
{
	FILE * fid ;
	char buf[2048] ;
	unsigned int i ;

	if (NULL == settings)
		return 1 ;

	sprintf(buf,"%s.log",settings->pairs->filename) ;
	fid = fopen (buf,"w+t") ;
	if (NULL != fid)
	{
		// add testing information
		sprintf(buf, "final settings\n") ;
		fwrite (buf, sizeof(char), strchr(buf,'\0')-buf, fid) ;
		fprintf(fid,"kappa_o = %f\n",settings->kappa_o) ;
		if (0 == settings->ardon)
			fprintf(fid,"kappa = %f\n",settings->kappa_a) ;
		else 
		{
			for (i=0;i<settings->pairs->dimen;i++)
				fprintf(fid,"kappa%d = %f\n", i+1, settings->kappa[i]) ;				
		}
		fprintf(fid,"kappa_m = %f\n", settings->kappa_m) ;
		fprintf(fid,"noise variance = %f\n", settings->noisevar) ;
		fclose(fid) ;
	}
	if (1==settings->ardon)
	{
       		sprintf(buf,"%s.ard",settings->pairs->filename) ;
        	fid = fopen (buf,"w+t") ;
        	if (NULL != fid)
        	{
				// add testing information
				for (i=0;i<settings->pairs->dimen;i++)
					fprintf(fid,"%.10f\n", settings->kappa[i]) ;
			}
			fclose(fid) ;
	}
	return 0 ;
}

// end of gpor_predict.cpp

