#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "pref.h"
#include "neal_matrix.h"

#define UPPERFUNC (10.0)

int Compute_MAP_Weights (Pref_Settings * settings) 
{
	double sigma, z, n, phi ;
	unsigned int i=0 ;
	Data_Node * node ;
	Pref_Node * pref ;

	assert(NULL!=settings) ;
	assert(settings->noisevar>0) ;

	sigma = sqrt(2.0*settings->noisevar) ; // sigma^\star
	// initial weight

	node = settings->pairs->front ;
	while ( node != NULL )
	{
		if (NULL != node->weight)
			free(node->weight) ;
		node->weight = (double *) calloc(settings->pairs->classes,sizeof(double)) ;
		if (NULL == node->weight)
		{
			printf("out of memory at node->weight.\n") ;
			exit(1) ;
		}
		node = node->next ; 		
	}

	node = settings->pairs->front ;
	while (NULL != node)
	{
		pref = node->pref.front ;
		while (NULL != pref)
		{
			z = (settings->alpha+i)->p_cache[pref->u-1] - (settings->alpha+i)->p_cache[pref->v-1] ; // f(u_k)-f(v_k)
			z = z / sigma ;
			n = normpdf(z) ;
			phi = 0.5 + normcdf(0,z) ;			
			pref->z = z ;
			pref->n = n ;
			pref->phi = phi ;
			node->weight[pref->u-1]+=n/phi/sigma ;
			node->weight[pref->v-1]-=n/phi/sigma ;		
			pref = pref -> next ;
		}
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;
	return 0 ;
}

double Pref_MAP_Update(Pref_Settings * settings)
{
	double error = 0 ;
	double sigma ;
	double * f ;
	double * a ;
	double * lambda ;
	unsigned int i, j, k ;
	double z1, phi1, n1 ;
	//double func = 0 ;
	double step = 1.0 ;
	Data_Node * node ;
	Pref_Node * pref ;
	double * t1 ;
	double * df ;
	double err ;

	if (NULL == settings)
	{
		printf("Input pointer is NULL.\n") ;
		return  0 ;
	}
	if (settings->noisevar<=0)
		return 0 ;

	sigma = sqrt(2.0*settings->noisevar) ; // sigma^\star

	lambda = (double *)calloc(settings->pairs->count*settings->pairs->count
		*settings->pairs->classes*settings->pairs->classes,sizeof(double)) ; // Lambda	
	f = (double *)malloc(settings->pairs->count*settings->pairs->classes*sizeof(double)) ;
	t1 = (double *)calloc(settings->pairs->count*settings->pairs->classes,sizeof(double)) ;
	df = (double *)calloc(settings->pairs->count*settings->pairs->classes,sizeof(double)) ;
	a = (double *)calloc(settings->pairs->count*settings->pairs->classes,sizeof(double)) ;

	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0; j<settings->pairs->classes; j++)
		{
			f[j*settings->pairs->count+i] = (settings->alpha+i)->p_cache[j] ; // current f
		}
	}
	
	/*for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<settings->pairs->count;j++)
			printf("%.2f ", settings->invcov[i*settings->pairs->count+j]) ;
			printf("\n") ;
	}*/

	// prepare w_i
	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		// who is old function
		// update alpha_old
		pref = node->pref.front ;
		while (NULL != pref)
		{
			z1 = f[(pref->u-1)*settings->pairs->count+i] - f[(pref->v-1)*settings->pairs->count+i] ; // f(u_k)-f(v_k)
			z1 = z1 / sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			a[(pref->u-1)*settings->pairs->count+i]-=n1/phi1/sigma ;		
			a[(pref->v-1)*settings->pairs->count+i]+=n1/phi1/sigma ;

			lambda[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] += 
				( (n1*n1)/(phi1*phi1)+z1*n1/phi1)/sigma/sigma ;
			lambda[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] += 
				( (n1*n1)/(phi1*phi1)+z1*n1/phi1)/sigma/sigma ;
			lambda[((pref->u-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->v-1)*settings->pairs->count+i)] -= 
				( (n1*n1)/(phi1*phi1)+z1*n1/phi1)/sigma/sigma ;
			lambda[((pref->v-1)*settings->pairs->count+i)*settings->pairs->count*settings->pairs->classes
				+((pref->u-1)*settings->pairs->count+i)] -= 
				( (n1*n1)/(phi1*phi1)+z1*n1/phi1)/sigma/sigma ;
		
			pref = pref -> next ;
		}
		/*(settings->alpha+i)->alpha = (settings->alpha+i)->pair->epamp ;
		(settings->alpha+i)->pair->epmean = (settings->alpha+i)->pair->postmean ;
	
		n1 = 0 ;
		n2 = 0 ;
		z1 = 0 ;
		z2 = 0 ;
		phi1 = 1 ;
		phi2 = 0 ;

		if (1 == (settings->alpha+i)->pair->target)
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			dphi = phi1-phi2 ;
		}
		else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
		{
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;
			n2 = normpdf(z2) ;
			phi2 = 0.5 + normcdf(0,z2) ;
			dphi = phi1-phi2 ;
		}
		else
		{
			z1 = (settings->thresholds[(settings->alpha+i)->pair->target-1]-(settings->alpha+i)->pair->epmean)/sigma ;
			z2 = (settings->thresholds[(settings->alpha+i)->pair->target-2]-(settings->alpha+i)->pair->epmean)/sigma ;			
			n1 = normpdf(z1) ;
			n2 = normpdf(z2) ;
			dphi = normcdf(z2,z1);
		}

		if (dphi<EPS)
		{
			if (n1-n2<EPS)
			{
				func -= log(EPS) ;
				step = 0.01 ;		
			}
			else
			{
				func -= log(EPS) ;
				step = 0.1 ;
			}

			if (1 == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = -z1/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else if (settings->pairs->classes == (settings->alpha+i)->pair->target)
			{
				(settings->alpha+i)->pair->weight = z2/sigma ;
				(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
			}
			else
			{
				if ((n1-n2)>=0)
				{
					(settings->alpha+i)->pair->weight = -(z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
					//(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma 
					//	+ (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight
					//	- (z1*z1*exp(-0.5*z1*z1+0.5*z2*z2)-z2*z2)/(exp(-0.5*z1*z1+0.5*z2*z2)-1.0)/sigma/sigma ;
				}
				else
				{
					//printf("n1 < n2 ;\n") ;
					(settings->alpha+i)->pair->weight = 0 ;
					(settings->alpha+i)->pair->epmean = 0 ;
					(settings->alpha+i)->pair->postmean = 0 ;
					(settings->alpha+i)->pair->epinvvar = 1/sigma/sigma ;
				}
			}
#ifdef _ORDINAL_DEBUG
			//printf("%u approximation, w=%f, p=%f\n",i,(settings->alpha+i)->pair->weight,(settings->alpha+i)->pair->epinvvar) ;
#endif
		}
		else
		{
			(settings->alpha+i)->pair->weight = (n1-n2)/(dphi)/sigma ;
			(settings->alpha+i)->pair->epinvvar = (settings->alpha+i)->pair->weight*(settings->alpha+i)->pair->weight 
				+ (z1*n1-z2*n2)/(dphi)/sigma/sigma ;
			// update func 			
			func -= log(dphi) ;
		}

		(settings->alpha+i)->pair->weight = -(settings->alpha+i)->pair->weight ;

		if ((settings->alpha+i)->pair->epinvvar>1.0/sigma/sigma)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = 1.0 - EPS*EPS ;
		}
		
		if ((settings->alpha+i)->pair->epinvvar<0)
		{
			//printf("error : %u, %f\n", i, (settings->alpha+i)->pair->epinvvar) ;
			(settings->alpha+i)->pair->epinvvar = EPS*EPS ;
		}*/
		node = node -> next ;
		i+=1 ;
	}
	assert(i==settings->pairs->count) ;

	for (k=0;k<settings->pairs->classes;k++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
			{
				t1[k*settings->pairs->count+i] += settings->invcov[i*settings->pairs->count+j]*f[k*settings->pairs->count+j] ;			
				t1[k*settings->pairs->count+j] += settings->invcov[j*settings->pairs->count+i]*f[k*settings->pairs->count+i] ;	
			}
			t1[k*settings->pairs->count+i] += settings->invcov[i*settings->pairs->count+i]*f[k*settings->pairs->count+i] ;	
		}
	}
	for (i=0;i<settings->pairs->count*settings->pairs->classes;i++)
	{
		a[i] += t1[i] ;
		//printf("delta a(%d)=%f\n",i,a[i]) ;
	}

	// Sigma-1+Lambda
	for (k=0;k<settings->pairs->classes;k++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
			{
				assert(settings->invcov[i*settings->pairs->count+j]==settings->invcov[j*settings->pairs->count+i]) ;
				lambda[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
					+i*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+j] 
					+= settings->invcov[i*settings->pairs->count+j] ; 
				lambda[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
					+j*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+i] 
					+= settings->invcov[j*settings->pairs->count+i] ; 
			}			
			lambda[k*settings->pairs->count*settings->pairs->count*settings->pairs->classes
				+i*settings->pairs->classes*settings->pairs->count+k*settings->pairs->count+i] 
				+= settings->invcov[i*settings->pairs->count+i] ; 		
		}
	}
	conjugate_gradient( lambda, a, df, settings->pairs->count*settings->pairs->classes ) ;

/*
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;

	for (i=0;i<settings->pairs->count;i++)
	{
		t1[i] = 0 ;
		for (j=0;j<i;j++)
		{
			t1[i] += settings->invcov[i*settings->pairs->count+j]*a[j] ;			
			t1[j] += settings->invcov[j*settings->pairs->count+i]*a[i] ;	
		}
		t1[i] += settings->invcov[i*settings->pairs->count+i]*a[i] ;	
	}
*/
	err = 0 ;
	for (i=0;i<settings->pairs->count;i++)
		err += df[i] * df[i] ;

	if (err/(double)settings->pairs->count>1.0) 
		step = (double)settings->pairs->count / err ;

	j=0 ;
	for (i=0;i<settings->pairs->count;i++)
	{
		for (j=0;j<settings->pairs->classes;j++)
		{
			(settings->alpha+i)->p_cache[j] -= step*df[j*settings->pairs->count+i] ;
			if (error<fabs(df[j]))		
				error = fabs(df[j]) ;
		}
		//printf("delta f(%d)=%f\n",i, t1[j]) ;	
		/*if (fabs((settings->alpha+j)->pair->postmean)>UPPERFUNC)
		{
//#ifdef _ORDINAL_DEBUG
			printf("%u - new %f --- old %f ---- w %f \n",i, (settings->alpha+j)->pair->postmean,(settings->alpha+j)->pair->epmean,(settings->alpha+j)->pair->weight) ;
//#endif
			//step = (UPPERFUNC-fabs((settings->alpha+i)->pair->epmean))/fabs(t1[i]) ;
			(settings->alpha+j)->pair->postmean = UPPERFUNC ;			
		}*/
	}

#ifdef _PREF_DEBUG
	
	i=0 ;
	node = settings->pairs->front ;
	while (NULL != node)
	{
		
		pref = node->pref.front ;
		while (NULL != pref)
		{
			z1 = f[(pref->u-1)*settings->pairs->count+i] - f[(pref->v-1)*settings->pairs->count+i] ; // f(u_k)-f(v_k)
			z1 = z1 / sigma ;
			n1 = normpdf(z1) ;
			phi1 = 0.5 + normcdf(0,z1) ;
			func -= log(phi1) ;
			pref = pref -> next ;
		}
		// update alpha_old
		i+=1 ;
		node = node -> next ;
	}
	for (k=0;k<settings->pairs->classes;k++)
	{
		for (i=0;i<settings->pairs->count;i++)
		{
			for (j=0;j<i;j++)
			{
				func += settings->invcov[i*settings->pairs->count+j] * f[k*settings->pairs->count+i] * f[k*settings->pairs->count+j] ;
			}
			func += 0.5 * settings->invcov[i*settings->pairs->count+i] * f[k*settings->pairs->count+i] * f[k*settings->pairs->count+i] ;
		}
	}
	printf("FUNCTIONAL : %f \n", func) ;
#endif 

	free (df) ;	
	free (t1) ;		
	free (a) ;
	free (f) ;
	free (lambda) ;
	return error ;
}

int Labels_MAP_Training (Pref_Settings * settings) 
{
	double error ;
	int index = 0 ;
	double * t1 ;
	double * t2 ;
	unsigned int i, j ;
	double temp ;
	
	if (NULL == settings)
		return 0 ;

	// may refresh ALPHAS here 
	if (NULL != settings->alpha)
		Clear_Alphas(settings->alpha, settings->pairs) ;
	settings->alpha = Create_Alphas (settings) ;
	if (NULL == settings->alpha)
	{
		printf("fail to create ALPHAs.\n") ;
		exit(1) ;
	}

#ifdef _ORDINAL_DEBUG
	printf("Current Settings: \n") ;
	printf("Noise Variance: %.3f\n",settings->noisevar) ;
	printf("Kappa_O: %.3f\n",settings->kappa_o) ;
	if (0==settings->ardon)		
		printf("Kernel: %.3f\n",settings->kappa_a) ;
	else
	{
		printf("Kernel:\n") ;
		for (i=0;i<settings->pairs->dimen;i++)
			printf("%u ---  %.3f:\n", i, settings->kappa[i]) ;
	}
#endif

	// intialize function
	// initialize the inverse matrix 
	t1 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	t2 = (double *)malloc(settings->pairs->count*sizeof(double)) ;
	if (NULL != settings->invcov) 
		free (settings->invcov) ;
	settings->invcov = (double *)malloc(settings->pairs->count*settings->pairs->count*sizeof(double)) ; // Sigma
	for (i=0;i<settings->pairs->count;i++)
	{			
		for (j=0;j<i;j++)
		{
			settings->invcov[i*settings->pairs->count+j] = 
				Calculate_Covfun(settings->alpha+i,settings->alpha+j,settings) ;
			settings->invcov[j*settings->pairs->count+i] = 
				settings->invcov[i*settings->pairs->count+j] ;
		}
		settings->invcov[i*settings->pairs->count+i] = Calculate_Covfun(settings->alpha+i,settings->alpha+i,settings) ;
	}
	if (0 == cholesky(settings->invcov,settings->pairs->count,&temp))
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	settings->trace = 0 ;
	for (i=0;i<settings->pairs->count;i++)
		settings->trace += log(settings->invcov[i*settings->pairs->count+i]) ;
	if (0 == inverse_from_cholesky(settings->invcov, t1, t2, settings->pairs->count) )		
		printf("Fatal Error: Sigma+1/Pi is not positive definite.\n") ;
	free(t1) ;
	free(t2) ;

	do
	{
		error = Pref_MAP_Update(settings) ; // clear sign
		index += 1 ;
#ifdef _ORDINAL_DEBUG
		printf("Loop %d --- %f\n", index, error) ;
#endif
	}
	while (error>TOL*TOL&&index<100) ;

	if (1 == Compute_MAP_Weights (settings))
		return 1 ;

#ifdef _ORDINAL_DEBUG
	/*for (index=0;index<(int)settings->pairs->count;index++)
		printf("%u --- post m %.3f - weight %.3f -alpha %.3f \n", 
		(settings->alpha+index)->pair->target, 
		(settings->alpha+index)->pair->postmean,
		(settings->alpha+index)->pair->weight,
		(settings->alpha+index)->alpha) ;*/
#endif
	return 0 ;
}



